// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; indent-tabs-mode: nil; -*-
//////////////////////////////////////////////////////////////////////
// File: timedgate.cpp
// This file contains sample implementation for the test generation:
// Timed train gate controller implementation under test.
// Authors:
//   Marius Mikucionis marius@cs.aau.dk
//////////////////////////////////////////////////////////////////////

#include "timedgate.h"
#include "tron/timing.h"
#include "tron/timeutil.h"

#include "assert.h"
#include <iostream>
#include <stdio.h>
#include <sys/errno.h>
#include <string.h>
#include <stdlib.h>
#include <queue>

using std::cout; // for stateDump
using std::cerr;
using std::endl;
using std::queue;


int instances=0;
extern "C" {
    TestAdapter* adapter_new(Reporter* rep, int argc, const char* args[]){
        if (instances>0) {
            cerr << "Only one instance at a time is supported. "
                "Destroy the old adapter and retry" << endl;
            return NULL;
        } else instances++;
        int m = 0;
        if (argc>1) m = atoi(args[1]);
        if (0<=m && m<=6) return new GateAdapter(rep, 4, m);
        else {
            cerr << "Mutant "<< m << " ("<<args[1]<<") not supported" << endl;
            return NULL;
        }
    };
    void adapter_delete(TestAdapter* ptr){ delete ptr; instances--; };
    void adapter_start(TestAdapter* a){ ((GateAdapter*)a)->start_imp(); };
    void adapter_perform(TestAdapter* a,
                         int32_t chan, uint16_t n, const int32_t data[]) {
        ((GateAdapter*)a)->perform_imp(chan, n, data);
    }
}


const uint16_t inCount = 2, outCount = 2;
const char* const inStr[inCount]={"appr", "leave"};
const char* const outStr[outCount]={"go", "stop"};
int32_t ins[inCount], outs[outCount], release=-1;
uint16_t appr=0, leave=1, go=0, stop=1;
int32_t timeUnit=10000;

queue<int32_t> inpChan; // input buffer for channel
queue<int32_t> inpId; // input buffer for gate id

pthread_mutex_t input_m;
pthread_cond_t input_c;
pthread_cond_t ack_c; // for input consumption notification

Gate::Gate(int i, GateAdapter *gateAdapter):
    Executable(), id(i), ga(gateAdapter), state(SID_FREE),
    tqueue(ga->tqueue), abort(false)
{
    tron_cond_init(&dead_c, NULL);
}

Gate::~Gate()
{
    tron_mutex_lock(&input_m);
    abort = true;
    tron_cond_broadcast(&input_c);
    tron_cond_broadcast(&ack_c);
    tron_cond_wait(&dead_c, &input_m);
    tron_mutex_unlock(&input_m);
    joinThread(NULL); // wait for completion
    tron_cond_destroy(&dead_c);
}

inline void Gate::start() // assume adapter_m is locked
{
    tron_mutex_lock(&input_m);
    startonce();
    tron_cond_wait(&ack_c, &input_m);// wait for start up
    tron_mutex_unlock(&input_m);
}

bool Gate::sendOutputOn(struct timespec& timeout,
                        int32_t chan, int16_t n, const int32_t data[])
{
    int res = 0;
    while (!abort) {
        if (!inpChan.empty()) { // there is some input
            if (inpId.front()==id) return false; // for me
            else tron_cond_wait(&ack_c, &input_m); //wait for input consumption
        } else {// there is no input
            if (res != ETIMEDOUT) // it's not timeout yet, let's wait
                res = tron_cond_timedwait(&input_c, &input_m, &timeout);
            if (abort) break;
            if  (!inpChan.empty()) {// input is still empty
                if (res == ETIMEDOUT) {// it's a timeout
                    if (chan==0) return true;// for mutants
                    else {
                        ga->reportOutput(chan, n, data);
                        return true;
                    }
                } else res = tron_cond_timedwait(&input_c, &input_m, &timeout);
            } else {
                if (inpId.front()==id) return false; // for me
                else tron_cond_wait(&ack_c, &input_m); //for input consumption
            }
        }
    }
    return false;
}

void Gate::state_free()
{
    while (!abort && (inpId.front()!=id))
        tron_cond_wait(&input_c, &input_m);
    if (abort) return;
    if (inpChan.front()==ins[appr]) state = SID_APPR;
    else cerr <<"GC"<<id<<": unacceptable input: "<<inpChan.front()<<endl;
    inpChan.pop(); inpId.pop();
    if (!inpChan.empty()) tron_cond_broadcast(&input_c);
}

void Gate::state_appr()
{
    if (tqueue->add(id)>1) state = SID_RED;
    else state = SID_WAIT;
}

void Gate::state_red()
{
    tron_gettime(&timeout); timeout += 1*timeUnit;
    if (sendOutputOn(timeout, outs[stop], 1, &id)) state = SID_HOLD;
    else { // there was some input for me
        if (abort) return;
        if (inpChan.front()==release) state = SID_WAIT;
        else cerr <<"GC"<<id<<": unacceptable input: "<<inpChan.front()<<endl;
        inpChan.pop(); inpId.pop();
        if (!inpChan.empty()) tron_cond_broadcast(&input_c);
    }
}

void Gate::state_hold()
{
    while (!abort && (inpId.front()!=id))
        tron_cond_wait(&input_c, &input_m);
    if (abort) return;
    if (inpChan.front()==release) state = SID_GREEN;
    else cerr <<"GC"<<id<<": unacceptable input: "<<inpChan.front()<<endl;
    inpChan.pop(); inpId.pop();
    if (!inpChan.empty()) tron_cond_broadcast(&input_c);
}

void Gate::state_green()
{
    tron_gettime(&timeout); timeout += 1*timeUnit;
    if (sendOutputOn(timeout, outs[go], 1, &id)) state = SID_WAIT;
    else { // there was some input for me
        if (abort) return;
        else cerr <<"GC"<<id<<": unacceptable input: "<<inpChan.front()<<endl;
        inpChan.pop(); inpId.pop();
        if (!inpChan.empty()) tron_cond_broadcast(&input_c);
    }
}

void Gate::state_wait()
{
    while (!abort && (inpId.front()!=id))
        tron_cond_wait(&input_c, &input_m);
    if (abort) return;
    if (inpChan.front()==ins[leave]) state = SID_REMOVE;
    else cerr <<"GC"<<id<<": unacceptable input: "<<inpChan.front()<<endl;
    inpChan.pop(); inpId.pop();
    if (!inpChan.empty()) tron_cond_broadcast(&input_c);
}

void Gate::state_remove()
{
    tqueue->rem(id); state = SID_FREE;
}

void* Gate::threadExecute()
{
    tron_mutex_lock(&input_m);
    tron_cond_broadcast(&ack_c);

    while (!abort) switch(state){
    case SID_FREE:  state_free();  break;
    case SID_APPR:  state_appr();  break;
    case SID_RED:   state_red();   break;
    case SID_HOLD:  state_hold();  break;
    case SID_GREEN: state_green(); break;
    case SID_WAIT:  state_wait();  break;
    case SID_REMOVE:state_remove();break;
    }
    tron_cond_broadcast(&dead_c);
    tron_mutex_unlock(&input_m);
    return NULL;
}

void Gate::dumpState()
{
    cout << "GC" << id;
    switch (state) {
    case SID_FREE:  cout << "(FREE)"; break;
    case SID_APPR:  cout << "(APPR)"; break;
    case SID_RED:   cout << "(RED)"; break;
    case SID_HOLD:  cout << "(HOLD)"; break;
    case SID_GREEN: cout << "(GREEN)"; break;
    case SID_WAIT:  cout << "(WAIT)"; break;
    case SID_REMOVE:cout << "(REMOVE)"; break;
    default: assert(false); break;
    }
}

TrainQueue::TrainQueue(GateAdapter *ta):
    head(NULL), tail(NULL), size(0), adapter(ta)
{
    if (0 != tron_mutex_init(&list_m, NULL/*use defaults*/)) {
        cerr << "TrainQueue: error while init mutex" << endl;
        exit(EXIT_FAILURE);
    }
}

TrainQueue::~TrainQueue()
{
    tron_mutex_lock(&list_m);
    while (head) {
        node* tmp = head;
        head = head->next;
        delete tmp;
    }
    tail = NULL;
    tron_mutex_unlock(&list_m);
    tron_mutex_destroy(&list_m);
}

int TrainQueue::add(int id)
{
    tron_mutex_lock(&list_m);
    if (size==0) {
        assert(!head && !tail);
        head = tail = new node(id);
    } else tail = tail->next = new node(id);
    size++;
    tron_mutex_unlock(&list_m);
    return size;
}

void TrainQueue::rem(int id)
{
    tron_mutex_lock(&list_m);
    assert(head && head->id == id);
    node* tmp = head; head = head->next; delete tmp; size--; // pop_front
    if (head) {
        assert(head->id != id); // must avoid callback mutex deadlock
        assert(head->id>0 && head->id<=4);
        adapter->locked_perform(release, 1, &(head->id));//FIXME: potential thread race over head->id
    } else tail = NULL;
    tron_mutex_unlock(&list_m);
}

// mutants of the TrainQueue:
class TrainQueue4: public TrainQueue
{
public:
    TrainQueue4(GateAdapter *ta): TrainQueue(ta) {}
    virtual ~TrainQueue4();
    virtual int add(int id); // id >= 0, returns size
    virtual void rem(int id);
};
class TrainQueue5: public TrainQueue
{
public:
    TrainQueue5(GateAdapter *ta): TrainQueue(ta) {}
    virtual ~TrainQueue5();
    virtual int add(int id); // id >= 0, returns size
    virtual void rem(int id);
};

//mutants of the Gate:
class Gate1: public Gate
{
    void state_red();
public:
    Gate1(int id, GateAdapter* ga): Gate(id, ga){}
};
class Gate2: public Gate
{
    int32_t bad_id;
    void state_red();
public:
    Gate2(int id, GateAdapter* ga): Gate(id, ga), bad_id(1) {}
};
class Gate3: public Gate
{
    void state_red();
public:
    Gate3(int id, GateAdapter* ga): Gate(id, ga){}
};
class Gate6: public Gate
{
    struct timespec busystart;
    void state_free();
public:
    Gate6(int id, GateAdapter* ga): Gate(id, ga){}
};


GateAdapter::GateAdapter(Reporter* rep, int railways, int mutant):
    TestAdapter(rep),
    gatecnt(railways), gates(new Gate*[railways]), tqueue(NULL)
{
    start = adapter_start;
    perform = adapter_perform;
    assert(gatecnt>0);
    if (0 != tron_mutex_init(&input_m, NULL/*use defaults*/)) {
        cerr << "GA: error while init mutex" << endl;
        exit(1);
    }
    tron_cond_init(&input_c, NULL);
    tron_cond_init(&ack_c, NULL);

    try { // initialize testing interface
        int r = 0;
        for (int i = 0; i < inCount; i++) {
            r = rep->getInputEncoding(rep, inStr[i]);
            if (r<0) throw r;
            ins[i] = r;
        }
        for (int i = 0; i < outCount; i++) {
            r = rep->getOutputEncoding(rep, outStr[i]);
            if (r<0) throw r;
            outs[i] = r;
        }
        r = rep->addVarToInput(rep, ins[appr], "inp"); if (r<0) throw r;
        r = rep->addVarToInput(rep, ins[leave], "inp"); if (r<0) throw r;
        r = rep->addVarToOutput(rep, outs[go], "out"); if (r<0) throw r;
        r = rep->addVarToOutput(rep, outs[stop], "out"); if (r<0) throw r;
        r = rep->setTimeUnit(rep, timeUnit); if (r<0) throw r;
        r = rep->setTimeout(rep, 100000); if (r<0) throw r;
    } catch (int error) {
        cerr << "GA: "<< rep->getErrorMessage(rep, error) << endl;
        exit(EXIT_FAILURE);
    }

    switch (mutant) {
    case 0:
    case 1:
    case 2:
    case 3:
    case 6: tqueue = new TrainQueue(this);  break;
    case 4: tqueue = new TrainQueue4(this); break;
    case 5: tqueue = new TrainQueue5(this); break;
    default: assert(false);
    }

    for (int i = 0; i < gatecnt; i++) {
        if (mutant==0 || mutant==4 || mutant==5)
            gates[i] = new Gate(i+1, this);
        else if (i != 3) gates[i]=new Gate(i+1, this);
        else switch(mutant) {
        case 1: gates[i] = new Gate1(i+1, this); break;
        case 2: gates[i] = new Gate2(i+1, this); break;
        case 3: gates[i] = new Gate3(i+1, this); break;
        case 6: gates[i] = new Gate6(i+1, this); break;
        default: assert(false);
        }
    }
}

GateAdapter::~GateAdapter()
{
    for (int i = 0; i < gatecnt; i++)
        delete gates[i];
    delete [] gates;
    delete tqueue;
    tron_cond_destroy(&input_c);
    tron_cond_destroy(&ack_c);
    tron_mutex_destroy(&input_m);
}

inline void GateAdapter::start_imp()
{
    assert(gates);
    for (int i = 0; i < gatecnt; i++) {
        assert(gates[i]);
        gates[i]->start();
    }
#ifdef SIMTIME
    clock = gates[gatecnt-1]->getClock();
#endif
}

inline void GateAdapter::locked_perform(int32_t chan, uint16_t n,
                                 const int32_t data[])
{
    inpChan.push(chan);
    if (n>0) inpId.push(data[0]);
    else inpId.push(0);
    tron_cond_broadcast(&input_c);
}

inline void GateAdapter::perform_imp(int32_t chan,
                                     uint16_t n, const int32_t data[])
{
    tron_mutex_lock(&input_m);
    locked_perform(chan, n, data);
    tron_mutex_unlock(&input_m);
}

void* GateAdapter::threadExecute(){}

/* Mutants methods: */

void Gate1::state_red()
{
    tron_gettime(&timeout);
    timeout += 6*timeUnit; // mutate: send stop too late
    if (sendOutputOn(timeout, outs[stop], 1, &id)) state = SID_HOLD;
    else { // there was some input for me
        if (abort) return;
        if (inpChan.front()==release) state = SID_WAIT;
        else cerr <<"GC"<<id<<": unacceptable input: "<<inpChan.front()<<endl;
        inpChan.pop(); inpId.pop();
        if (!inpChan.empty()) tron_cond_broadcast(&input_c);
    }
}

void Gate2::state_red()
{
    tron_gettime(&timeout);
    timeout += 1*timeUnit;
    if (sendOutputOn(timeout, outs[stop], 1, &bad_id)) //mutate: bad id
        state = SID_HOLD;
    else { // there was some input for me
        if (abort) return;
        if (inpChan.front()==release) state = SID_WAIT;
        else cerr <<"GC"<<id<<": unacceptable input: "<<inpChan.front()<<endl;
        inpChan.pop(); inpId.pop();
        if (!inpChan.empty()) tron_cond_broadcast(&input_c);
    }
}

void Gate3::state_red()
{
    tron_gettime(&timeout); timeout += 1*timeUnit;
    if (sendOutputOn(timeout, 0, 0, &id)) {// mutate: loose stop
        state = SID_HOLD;
    } else { // there was some input for me
        if (abort) return;
        if (inpChan.front()==release) state = SID_WAIT;
        else cerr <<"GC"<<id<<": unacceptable input: "<<inpChan.front()<<endl;
        inpChan.pop(); inpId.pop();
        if (!inpChan.empty()) tron_cond_broadcast(&input_c);
    }
}

void Gate6::state_free()
{
    tron_gettime(&busystart);// mutate: remember the start moment
    while (!abort && (inpId.front()!=id))
        tron_cond_wait(&input_c, &input_m);
    if (abort) return;
    if (inpChan.front()==ins[appr]) {
        tron_gettime(&timeout);// mutate: simulate busy for 2 time units.
        if (diff2ms(timeout, busystart)>=2*timeUnit) state = SID_APPR;
    } else cerr <<"GC"<<id<<": unacceptable input: "<<inpChan.front()<<endl;
    inpChan.pop(); inpId.pop();
    if (!inpChan.empty()) tron_cond_broadcast(&input_c);
}

int TrainQueue4::add(int id)
{
    tron_mutex_lock(&list_m);
    if (size==0) {
        assert(!head && !tail);
        head = tail = new node(id);
        size++;
    } else if (size < 3) {
        tail = tail->next = new node(id);
        size++;
    } else { // mutate: allow only bounded size
        tail->id = id;
    }
    tron_mutex_unlock(&list_m);
    return size;
}

void TrainQueue4::rem(int id)
{
    tron_mutex_lock(&list_m);
    assert(head && head->id == id);
    if (size>0) {
        node* tmp = head; head = head->next; delete tmp; size--; // pop_front
    }
    if (size>0) {
        assert(head->id != id); // must avoid callback mutex deadlock
        assert(head->id>0 && head->id<=4);
        adapter->locked_perform(release, 1, &(head->id));
    } else tail = NULL;
    tron_mutex_unlock(&list_m);
}

TrainQueue4::~TrainQueue4(){}


int TrainQueue5::add(int id)
{
    tron_mutex_lock(&list_m);
    if (size==0) {
        assert(!head && !tail);
        head = tail = new node(id);
    } else tail = tail->next = new node(id);
    size++;
    tron_mutex_unlock(&list_m);
    return size;
}

void TrainQueue5::rem(int id)// LIFO instead of FIFO
{
    tron_mutex_lock(&list_m);
    assert(head);
    node *tmp = head;
    if (tmp->id == id) { // find and remove the element
        head = head->next;
        delete tmp; size--;
    } else {
        while (tmp->next->id != id) tmp=tmp->next;
        node *tmp2 = tmp->next;
        tmp->next = tmp2->next;
        delete tmp2; size--;
    }
    if (size>0) { // notify someone at the tail
        assert(tail->id != id); // must avoid callback mutex deadlock
        assert(tail->id>0 && tail->id<=4);
        adapter->locked_perform(release, 1, &(tail->id));
    } else tail = NULL;
    tron_mutex_unlock(&list_m);
}

TrainQueue5::~TrainQueue5(){}
